/*
	$Id: connection.c,v 1.11 2001/09/22 14:57:12 syl Exp $

	Copyright (C) 1999, 2000 Syllabub
	Maroontress Fast Software.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/Xresource.h>
#include <stdlib.h>

#include "Resource.h"
#include "Text.h"
#include "phrase.h"
#include "connection.h"
#include "leakdetect.h"
#include "font.h"

InputContext *
CreateInputContext(Connection *con)
{
    InputContext *ic;

    if (con->idle == NULL) {
	if (con->next_ic_id > 0xffff)
	    return (NULL);
	if ((ic = (InputContext *)ldmalloc(sizeof(InputContext))) == NULL)
	    return (NULL);
	ic->ic_id = con->next_ic_id;
	++(con->next_ic_id);
    }
    else {
	ic = con->idle;
	con->idle = ic->next;
    }
    ic->next = con->work;
    ic->input_style = 0;
    ic->client_window = None;
    ic->focus_window = None;
    ic->preedit_window = None;
    ic->preedit_fontset.id = NULL;
    ic->preedit_fontset_name = NULL;
    ic->preedit_foreground = BlackPixel(con->disp, DefaultScreen(con->disp));
    ic->preedit_background = WhitePixel(con->disp, DefaultScreen(con->disp));
    ic->preedit_spot_location.x = 0;
    ic->preedit_spot_location.y = 0;
    ic->status_window = None;
    ic->status_pixmap = None;
    ic->status_gc = None;
    ic->status_fontset.id = NULL;
    ic->status_fontset_name = NULL;
    ic->status_foreground = BlackPixel(con->disp, DefaultScreen(con->disp));
    ic->status_background = WhitePixel(con->disp, DefaultScreen(con->disp));
    ic->status_area.x = 0;
    ic->status_area.y = 0;
    ic->status_area.width = 0;
    ic->status_area.height = 0;
    ic->candidate_window = None;
    ic->candidate_fontset = NULL;
    ic->forward = False;
    ic->focus = False;
    ic->preconversion = PRECONVERSION_NONE;
    ic->max_preedit_chars = 0;
    ic->preedit = CreateSylTextFromMBString("", False);
    ic->fixed = CreateSylTextFromMBString("", False);
    ic->caret = 0;
    ic->head = NULL;
    ic->view = NULL;
    ic->preedit_draw = NULL;
    ic->status_draw = NULL;
    ic->nice_next = NULL;
    ic->nice_func = NULL;
    ic->nice_data = NULL;
    con->work = ic;
    return (ic);
}

InputContext *
SerachInputContext(Connection *con, int ic_id)
{
    InputContext *p;

    for (p = con->work; p != NULL && p->ic_id != ic_id; p = p->next)
	;
    return (p);
}

static void
DestroyInputContext(Connection *con, InputContext *ic, int destroyed)
{
    ic->preedit_draw->destroy(con, ic);
    ic->status_draw->destroy(con, ic);

    if (ic->preedit_window != None)
	XDeleteContext(con->disp, ic->preedit_window, con->preedit);
    if (ic->preedit_window != None && !destroyed)
	XDestroyWindow(con->disp, ic->preedit_window);
    if (ic->preedit_fontset.id != NULL)
	FreeSharedFontset(con->disp, &ic->preedit_fontset);
    if (ic->preedit_fontset_name != NULL)
	free(ic->preedit_fontset_name);

    if (ic->status_fontset.id != NULL)
	FreeSharedFontset(con->disp, &ic->status_fontset);
    if (ic->status_fontset_name != NULL)
	free(ic->status_fontset_name);
    if (ic->status_window != None)
	XDeleteContext(con->disp, ic->status_window, con->status);
    if (ic->status_window != None && !destroyed)
	XDestroyWindow(con->disp, ic->status_window);
    if (ic->status_pixmap != None)
	XFreePixmap(con->disp, ic->status_pixmap);
    if (ic->status_gc != None)
	XFreeGC(con->disp, ic->status_gc);

    if (ic->candidate_window != None) {
	XDeleteContext(con->disp, ic->candidate_window, con->candidate);
	XDestroyWindow(con->disp, ic->candidate_window);
    }
    FreeSylText(ic->preedit);
    FreeSylText(ic->fixed);
    FreePhrase(ic->head);
}

void
FreeInputContext(Connection *con, int ic_id)
{
    InputContext *ic, *p;

    if ((ic = SerachInputContext(con, ic_id)) == NULL)
	return;

    if (con->work == ic) {
	con->work = ic->next;
    }
    else {
	for (p = con->work; p->next != ic; p = p->next)
	    ;
	p->next = ic->next;
    }
    DestroyInputContext(con, ic, False);
    ic->next = con->idle;
    con->idle = ic;
}

static InputContext *Deferred = NULL;

void
DeferInputContext(InputContext *ic, void (*func)(InputContext *, void *),
		  void *data)
{
    if (ic->nice_func == NULL) {
	ic->nice_next = Deferred;
	Deferred = ic;
    }
    ic->nice_func = func;
    ic->nice_data = data;
}

int
NiceInputContext(void)
{
    InputContext *p = Deferred;
    void (*func)(InputContext *, void *);

    if (p == NULL)
	return (False);

    Deferred = p->nice_next;
    p->nice_next = NULL;
    func = p->nice_func;
    p->nice_func = NULL;
    func(p, p->nice_data);
    /* p may be no longer valid pointer */
    return (True);
}

static int NextID = 1;
static Connection *Work = NULL;
static Connection *Idle = NULL;

Connection *
CreateConnection(XClientMessageEvent *msg)
{
    Connection *con;

    if (Idle == NULL) {
	if (NextID > 0xffff)
	    return (NULL);
	if ((con = (Connection *)ldmalloc(sizeof(Connection))) == NULL)
	    return (NULL);
	con->im_id = NextID;
	++NextID;
    }
    else {
	con = Idle;
	Idle = con->next;
    }
    con->next = Work;
    con->disp = msg->display;
    con->client = (Window)msg->data.l[0];
    con->server = XCreateSimpleWindow(con->disp, con->client,
				      0, 0, 1, 1, 0, 0, 0);
    XSelectInput(con->disp, con->server, StructureNotifyMask);
    con->work = NULL;
    con->idle = NULL;
    con->next_ic_id = 1;
    Work = con;
    return (con);
}

static void
DestroyCB(InputContext *ic, void *data)
{
    Connection *con = (Connection *)data;
    
    DestroyInputContext(con, ic, True);
    ldfree(ic);
}

void
FreeConnection(Connection *con)
{
    InputContext *p, *next;
    Connection *q;

    for (p = con->work; p != NULL; p = next) {
	next = p->next;
	DeferInputContext(p, DestroyCB, (void *)con);
    }
    for (p = con->idle; p != NULL; p = next) {
	next = p->next;
	ldfree(p);
    }
    /* con->server˲롣 */
    if (Work == con) {
	Work = con->next;
    }
    else {
	for (q = Work; q->next != con; q = q->next)
	    ;
	q->next = con->next;
    }
    con->next = Idle;
    Idle = con;
}
